"""Module to plot predicition, etc.

This file regroups all functions used to plot the input data or the regression results
"""

import base64
import os
from io import BytesIO

import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import xycmap
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.ndimage import median_filter

mpl.use("agg")

CM_PER_INCH = 1 / 2.54


def convert_plt_to_base64(plt):
    """Convert matplot to base64 encoded image."""
    img_buffer = BytesIO()
    plt.savefig(img_buffer, format="svg", bbox_inches="tight")
    img_buffer.seek(0)
    content = base64.b64encode(img_buffer.read()).decode()
    return content


def set_axis(rf_prediction):
    """Define the axis for plotting."""
    return (
        rf_prediction.rect_geom.grid_x / 1000,
        rf_prediction.rect_geom.grid_y / 1000,
    )


def plot_selection(rf_prediction):
    """Plot funciton defined by the input selection."""
    xaxis, yaxis = set_axis(rf_prediction)
    output_dir = rf_prediction.working_dir

    if rf_prediction.what_to_plot["predictors"]:
        plot_predictors(
            xaxis, yaxis, rf_prediction.input_data.predictors, output_dir=output_dir
        )

    if rf_prediction.what_to_plot["pred_correlation"]:
        prediction_correlation_matrix(rf_prediction.input_data, output_dir=output_dir)

    for time_step in range(
        len(rf_prediction.input_data.soil_moisture_data.start_times)
    ):
        if rf_prediction.what_to_plot["day_measurements"]:
            plot_measurements(
                rf_prediction.input_data.soil_moisture_data,
                time_step,
                rf_prediction.apply_monte_carlo,
                output_dir=output_dir,
            )
        if rf_prediction.what_to_plot["day_prediction_map"]:
            plot_rf_prediction(
                xaxis,
                yaxis,
                rf_prediction.input_data.soil_moisture_data,
                rf_prediction.RF_prediction,
                rf_prediction.MC_mean,
                rf_prediction.dispersion_coefficient,
                time_step,
                rf_prediction.apply_monte_carlo,
                output_dir=output_dir,
            )
        if rf_prediction.what_to_plot["day_predictor_importance"]:
            plot_predictor_importance(
                rf_prediction.input_data.soil_moisture_data,
                time_step,
                rf_prediction.input_data.number_predictors,
                rf_prediction.apply_monte_carlo,
                rf_prediction.input_data.predictors.keys(),
                rf_prediction.predictor_importance,
                output_dir=output_dir,
            )
    if rf_prediction.what_to_plot["alldays_predictor_importance"]:
        predictor_importance_along_days(
            rf_prediction.input_data.number_predictors,
            rf_prediction.input_data.soil_moisture_data.start_times,
            rf_prediction.predictor_importance,
            rf_prediction.input_data.predictors,
            output_dir=output_dir,
        )
    

# TODO give only needed argument instead of rf_prediction
def plot_predictors(xaxis, yaxis, predictors, output_dir=None, return_base64_img=False):
    """Plot all predictors as color maps, with name and unit."""
    n_cols = np.ceil(len(predictors) / 2).astype(int)
    fig, ax = plt.subplots(
        nrows=2,
        ncols=n_cols,
        sharex=True,
        sharey=True,
        figsize=(17 * CM_PER_INCH, 9 * CM_PER_INCH),
    )
    fig.subplots_adjust(wspace=0.4)
    plt.rcParams.update({"font.size": 5})
    axes_count = 0
    for pred_name, predictor in predictors.items():
        im = ax.flat[axes_count].pcolormesh(xaxis, yaxis, predictor[0], shading="auto")
        ax.flat[axes_count].set_title(pred_name)
        ax.flat[axes_count].set_aspect(1)
        divider = make_axes_locatable(ax.flat[axes_count])
        cax = divider.append_axes("right", size="5%", pad=0.15)
        cbar = plt.colorbar(im, cax=cax)
        cbar.ax.set_title(predictor[1])
        axes_count += 1

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "predictors.svg")
    plt.savefig(fig_file_path, dpi=300)


def prediction_correlation_matrix(input_data, output_dir=None, return_base64_img=False):
    """Plot the correlation matrix between all predictors, 2 by 2."""
    ticks = list(input_data.predictors.keys())
    plt.figure(figsize=(12 * CM_PER_INCH, 9 * CM_PER_INCH))
    plt.rcParams.update({"font.size": 5})
    correlation_matrix = input_data.compute_correlation_matrix()
    sns.heatmap(
        correlation_matrix,
        annot=True,
        cmap="seismic",
        vmin=-1,
        vmax=1,
        xticklabels=ticks,
        yticklabels=ticks,
    )

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "correlation_matrix.svg")
    plt.savefig(fig_file_path, dpi=300)


def draw_error_band_path(x, y, error):
    """Calculate normals via centered finite differences.

    Except the first point which uses a forward difference and the last point which uses
    a backward difference.
    """
    dist_to_next_point_x_component = np.concatenate(
        [[x[1] - x[0]], x[2:] - x[:-2], [x[-1] - x[-2]]]
    )
    dist_to_next_point_y_component = np.concatenate(
        [[y[1] - y[0]], y[2:] - y[:-2], [y[-1] - y[-2]]]
    )
    dist_to_next_point = np.hypot(
        dist_to_next_point_x_component, dist_to_next_point_y_component
    )
    normal_x_component = dist_to_next_point_y_component / dist_to_next_point
    normal_y_component = -dist_to_next_point_x_component / dist_to_next_point

    scale_error_vector = 3
    x_error_end_point = x + normal_x_component * error * scale_error_vector
    y_error_end_point = y + normal_y_component * error * scale_error_vector

    vertices = np.block([[x_error_end_point, x[::-1]], [y_error_end_point, y[::-1]]]).T
    codes = np.full(len(vertices), Path.LINETO)
    codes[0] = Path.MOVETO
    return Path(vertices, codes)


def plot_measurements_vs_distance(
    soil_moisture_data, time_step, monte_carlo, output_dir=None, return_base64_img=False
):
    """Plot soil moisture measurements vs. cumulative driven distance.

    Includes the corresponding low and up standard deviations.
    """
    start_time = soil_moisture_data.start_times[time_step]

    plt.rcParams.update({"font.size": 7})
    cumulative_distance = [0]
    for measurement in range(soil_moisture_data.number_measurements[start_time] - 1):
        cumulative_distance.append(
            (
                cumulative_distance[measurement]
                + np.sqrt(
                    (
                        soil_moisture_data.x_measurement[start_time][measurement + 1]
                        - soil_moisture_data.x_measurement[start_time][measurement]
                    )
                    ** 2
                    + (
                        soil_moisture_data.y_measurement[start_time][measurement + 1]
                        - soil_moisture_data.y_measurement[start_time][measurement]
                    )
                    ** 2
                )
                / 1000
            )
        )
    plt.figure()
    plt.plot(
        cumulative_distance,
        soil_moisture_data.soil_moisture[start_time],
        "g",
        label="Measurements",
    )
    if monte_carlo:
        plt.fill_between(
            cumulative_distance,
            soil_moisture_data.soil_moisture[start_time]
            - soil_moisture_data.soil_moisture_dev_low[start_time],
            soil_moisture_data.soil_moisture[start_time]
            + soil_moisture_data.soil_moisture_dev_high[start_time],
            edgecolor="none",
            alpha=0.3,
            color="purple",
            label="Lower/upper SD",
        )
    plt.ylim([0.1, 0.8])
    plt.xlabel("Cumulative distance (km)")
    plt.ylabel("Gravimetric soil moisture (g/g)")
    plt.legend(loc="upper left")

    if return_base64_img:
        return convert_plt_to_base64(plt)
    fig_file_path = os.path.join(
        output_dir, "measurements_vs_distance_" + start_time + ".svg"
    )
    plt.savefig(fig_file_path, dpi=300)


def plot_measurements(
    soil_moisture_data, time_step, monte_carlo, output_dir=None, return_base64_img=False
):
    """Plot soil moisture measurements vs. cumulative driven distance.

    Includes the corresponding low and up standard deviations.
    """
    start_time = soil_moisture_data.start_times[time_step]

    plt.figure()
    plt.gca().set_aspect(1)
    x = soil_moisture_data.x_measurement[start_time] / 1000
    y = soil_moisture_data.y_measurement[start_time] / 1000
    sc = plt.scatter(
        x,
        y,
        c=soil_moisture_data.soil_moisture[start_time],
        cmap="Spectral",
        s=5,
        vmin=0.1,
        vmax=0.6,
        zorder=2,
        label="Measurements",
    )
    if monte_carlo:
        plt.gca().add_patch(
            PathPatch(
                draw_error_band_path(
                    x, y, -soil_moisture_data.soil_moisture_dev_low[start_time]
                ),
                alpha=0.3,
                color="purple",
                label="Lower/upper SD",
            )
        )
        plt.gca().add_patch(
            PathPatch(
                draw_error_band_path(
                    x, y, soil_moisture_data.soil_moisture_dev_high[start_time]
                ),
                alpha=0.3,
                color="purple",
            )
        )
    plt.xlabel("Easting (km)")
    plt.ylabel("Northing (km)")
    plt.legend(loc="upper left")
    cbar = plt.colorbar(sc, shrink=0.55)
    cbar.set_label("Gravimetric soil moisture (g/g)")

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "measurements_" + start_time + ".svg")
    plt.savefig(fig_file_path, dpi=300)


def plot_rf_prediction(
    xaxis,
    yaxis,
    soil_moisture_data,
    rf_prediction,
    mc_mean,
    dispersion_coefficient,
    time_step,
    monte_carlo,
    *args,
    **kwargs
):
    """Plot random forest prediction as a color map."""
    if monte_carlo:
        return plot_rf_prediction_with_dispersion(
            xaxis,
            yaxis,
            soil_moisture_data,
            mc_mean,
            dispersion_coefficient,
            time_step,
            *args,
            **kwargs
        )
    else:
        return plot_rf_prediction_no_dispersion(
            xaxis, yaxis, soil_moisture_data, rf_prediction, time_step, *args, **kwargs
        )


def plot_rf_prediction_no_dispersion(
    xaxis,
    yaxis,
    soil_moisture_data,
    rf_prediction,
    time_step,
    output_dir=None,
    return_base64_img=False,
):
    """Plot random forest prediction as a color map."""
    if not output_dir:
        output_dir = rf_prediction.working_dir

    start_time = soil_moisture_data.start_times[time_step]

    plt.rcParams.update({"font.size": 6})
    fig, ax = plt.subplots(figsize=(16 * CM_PER_INCH, 14 * CM_PER_INCH))
    im = plt.pcolormesh(
        xaxis,
        yaxis,
        rf_prediction[0, time_step, :, :],
        vmin=0.15,
        vmax=0.35,
        cmap="Spectral",
    )

    plt.scatter(
        soil_moisture_data.x_measurement[start_time] / 1000,
        soil_moisture_data.y_measurement[start_time] / 1000,
        c="black",
        s=0.5,
    )
    ax.set_title(start_time)
    ax.set_aspect(1)
    cbar = plt.colorbar(im)
    cbar.ax.tick_params(labelsize=14)
    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "RF_prediction_" + start_time + ".tif")
    plt.savefig(fig_file_path, dpi=300)


def plot_rf_prediction_with_dispersion(
    xaxis,
    yaxis,
    soil_moisture_data,
    mc_mean,
    dispersion_coefficient,
    time_step,
    output_dir=None,
    return_base64_img=False,
):
    """Plot SM mean prediction and coefficient of dispersion maps.

    Measurement locations are overlaid.
    """
    start_time = soil_moisture_data.start_times[time_step]

    plt.rcParams.update({"font.size": 7})

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(16 / 2.54, 10 / 2.54))
    axes[0].set_aspect(1)
    axes[1].set_aspect(1)

    divider = make_axes_locatable(axes[0])
    cax = divider.append_axes("right", size="5%", pad=0.15)
    im = axes.flat[0].pcolormesh(
        xaxis,
        yaxis,
        mc_mean[time_step],
        shading="auto",
        vmin=0.2,
        vmax=0.5,
        cmap="Spectral",
    )
    axes.flat[0].scatter(
        soil_moisture_data.x_measurement[start_time] / 1000,
        soil_moisture_data.y_measurement[start_time] / 1000,
        c="black",
        s=1,
    )
    axes.flat[0].set_title("Mean")
    plt.colorbar(im, cax=cax)

    im1 = axes.flat[1].pcolormesh(
        xaxis,
        yaxis,
        dispersion_coefficient[time_step],
        shading="auto",
        vmin=0,
        vmax=0.05,
        cmap="Reds",
    )
    axes.flat[1].set_title("Coefficient of dispersion")
    axes.flat[1].tick_params(axis="y", left=False, labelleft=False)
    divider = make_axes_locatable(axes[1])
    cax = divider.append_axes("right", size="5%", pad=0.15)
    plt.colorbar(im1, cax=cax)

    fig.suptitle(start_time)

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "RF_prediction_" + start_time + ".svg")
    plt.savefig(fig_file_path, dpi=300)


def plot_predictor_importance(
    soil_moisture_data,
    time_step,
    number_predictors,
    monte_carlo,
    predictors,
    predictor_importance,
    output_dir=None,
    return_base64_img=False,
):
    """Plot predictor importance from the RF model."""
    start_time = soil_moisture_data.start_times[time_step]

    plt.rcParams.update({"font.size": 7})
    x = np.arange(number_predictors)
    plt.figure(figsize=(16 / 2.54, 7 / 2.54))
    if monte_carlo:
        plt.bar(
            predictors,
            np.percentile(
                predictor_importance[:, time_step, :],
                95,
                axis=0,
            ),
            color="deepskyblue",
            label="q95",
        )
        plt.bar(
            predictors,
            np.percentile(
                predictor_importance[:, time_step, :],
                50,
                axis=0,
            ),
            color="blue",
            label="median",
        )
        plt.bar(
            predictors,
            np.percentile(
                predictor_importance[:, time_step, :],
                5,
                axis=0,
            ),
            color="navy",
            label="q5",
        )
        plt.legend()
    else:
        plt.bar(predictors, predictor_importance[0, time_step, :])
    plt.xticks(x, predictors)
    plt.title(start_time)

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(
        output_dir, "predictor_importance_" + start_time + ".svg"
    )
    plt.savefig(fig_file_path, dpi=300)


def predictor_importance_along_days(
    number_predictors,
    start_times,
    predictor_importance,
    predictors,
    apply_pca=False,
    output_dir=None,
    return_base64_img=False,
):
    """Plot predictor importance from the RF model.

    input_data     : input_data object containing the list of all predictors
    rf_prediction : rf_prediction object
    """
    plt.rcParams.update({"font.size": 7})
    fig, ax = plt.subplots(
        number_predictors,
        sharex=True,
        figsize=(17 / 2.54, 16 / 2.54),
    )
    if not apply_pca:
        if number_predictors == 1:
            ax.plot(
                range(len(start_times)),
                np.mean(predictor_importance[:, :, 0], axis=0),
                label=list(predictors.keys())[0],
            )
            ax.set_ylim(0, 1)
            ax.legend(loc="right")
        else:
            for pred_index, predictor_name in enumerate(predictors.keys()):
                ax[pred_index].plot(
                    range(len(start_times)),
                    np.mean(predictor_importance[:, :, pred_index], axis=0),
                    label=predictor_name,
                )
                ax[pred_index].set_ylim(0, 1)
                ax[pred_index].legend(loc="right")
    else:
        for predictor_index in range(number_predictors):
            ax[predictor_index].plot(
                range(len(start_times)),
                np.mean(predictor_importance[:, :, predictor_index], axis=0),
                label="PCA comp. " + str(predictor_index),
            )
            ax[predictor_index].set_ylim(0, 1)
            ax[predictor_index].legend(loc="right")

    if return_base64_img:
        return convert_plt_to_base64(plt)

    fig_file_path = os.path.join(output_dir, "predictor_importance_vs_days.svg")
    plt.savefig(fig_file_path, dpi=300)


# Update!
def plot_section_sm_with_pred(rf_prediction, x_axis, y, pred, day):
    """Plot SM mean together with predictor values along this section.

    Plot SM mean prediction and percentiles along a horizontal or vertical map section,
    together with predictor values along this section

    rf_prediction : rf_prediction object
    x_axis        : 1D array with the coordinates along the section
    y             : scalar, fix coordinate of the section
    pred          : dictionary with predictor values along this section
    day             : int, day number
    """
    raise NotImplementedError

    rf_stats = [
        rf_prediction.p5[day],
        rf_prediction.p25[day],
        rf_prediction.m[day],
        rf_prediction.p75[day],
        rf_prediction.p95[day],
    ]
    p5_mid, p25_mid, p50_mid, p75_mid, p95_mid = [
        np.empty(len(x_axis)) for k in range(5)
    ]
    for j in range(len(x_axis)):
        idx_x = j
        p5_mid[j], p25_mid[j], p50_mid[j], p75_mid[j], p95_mid[j] = [
            perc[idx_x, y] for perc in rf_stats
        ]

    plt.rcParams.update({"font.size": 14})
    fig, ax1 = plt.subplots(figsize=(23, 5))
    ax1.plot(x_axis, p50_mid, "r", label="Median")
    ax1.fill_between(x_axis, p5_mid, p95_mid, color="blue", alpha=0.3, label="5-95%")
    ax1.fill_between(x_axis, p25_mid, p75_mid, color="blue", alpha=0.7, label="25-75%")
    ax1.set_ylim([0.2, 0.5])
    ax1.set_ylabel("Gravimetric soil moisture (g/g)")
    ax2 = ax1.twinx()
    for k, v in pred.items():
        ax2.plot(x_axis, v[0][:, y], label=k)
    ax2.legend()


# Update!
def plot_section_all_iters(rf_prediction, x_axis, y, day):
    """Plot all SM RF predictions along a horizontal or vertical map section.

    rf_prediction : rf_prediction object
    x_axis        : 1D array with the coordinates along the section
    y             : scalar, fix coordinate of the section
    day             : int, day number
    """
    raise NotImplementedError
    plt.rcParams.update({"font.size": 14})
    plt.figure()
    plt.ylim([0.2, 0.5])
    for i in range(rf_prediction.n_iters):
        plt.plot(x_axis, rf_prediction.RF_prediction[i, day, :, y])


def plot_rf_prediction_as_bivariate(rf_prediction, day, smooth_data=False):
    """Create a bivariate map based on a 2D colormap (saturation of a given colormap).

    rf_prediction : rf_prediction object
    """
    if smooth_data:
        mc_mean = median_filter(rf_prediction.MC_mean[day], size=3, mode="nearest")
        disp_coeff = median_filter(
            rf_prediction.disp_coeff[day], size=3, mode="nearest"
        )

    cmap = saturation_xycmap(mpl.cm.get_cmap("Spectral"), "white", n=(9, 3))
    colors = xycmap.bivariate_color(
        sx=mc_mean.flatten(),
        sy=disp_coeff.flatten(),
        cmap=cmap,
        xlims=(0.2, 0.5),
        ylims=(0.2, 0.55),
    )
    color = colors.to_numpy().reshape(mc_mean.shape).T
    bi_map = np.empty((color.shape[0], color.shape[1], 4))  # 4 for rgba
    for i in range(color.shape[0]):
        for j in range(color.shape[1]):
            bi_map[i, j, 0] = color[i, j][0]
            bi_map[i, j, 1] = color[i, j][1]
            bi_map[i, j, 2] = color[i, j][2]
            bi_map[i, j, 3] = color[i, j][3]
    plt.rcParams.update({"font.size": 13})
    fig, axs = plt.subplots(1, 2)
    axs[0].imshow(
        bi_map,
        origin="lower",
        extent=[
            rf_prediction.geom.xi / 1000,
            rf_prediction.geom.xf / 1000,
            rf_prediction.geom.yi / 1000,
            rf_prediction.geom.yf / 1000,
        ],
    )
    axs[0].scatter(
        rf_prediction.input_data.x_crns / 1000,
        rf_prediction.input_data.y_crns / 1000,
        c="black",
        s=1,
    )
    axs[0].set_ylabel("Northing (km)")
    axs[0].set_xlabel("Easting (km)")
    xycmap.bivariate_legend(
        ax=axs[1],
        sx=mc_mean.flatten(),
        sy=disp_coeff.flatten(),
        cmap=cmap,
        xlims=(0.2, 0.5),
        ylims=(0.2, 0.55),
    )
    axs[1].set_xlabel("Gravimetric soil moisture (g/g)")
    axs[1].set_ylabel("Coefficient of dispersion")


def saturation_xycmap(xcmap, color, n):
    """Create a 2D colormap by saturation of a given colormap.

    Args:
        xcmap: Matplotlib colormap along the x-axis.
        color: Color used for the colormap saturation along the y-axis.
        n    : Tuple containing the number of columns and rows (x, y).
    Returns:
        Custom two-dimensional colormap in np.ndarray.
    Raises:
        ValueError: If less than two columns or rows are passed.
    """
    xn, yn = n
    if xn < 2 or yn < 2:
        raise ValueError("Expected n >= 2 categories.")

    sy, sx = np.mgrid[0:yn, 0:xn]
    scale = sy / (yn - 1)

    xvals = np.array(
        255 * (sx - sx.min()) / (sx.max() - sx.min()), dtype=int
    )  # Rescale the mock series into the colormap range (0, 255).

    xcolors = np.asarray(xcmap(xvals))
    ycolors = np.empty((yn, xn, 4))
    for i in range(yn):
        for j in range(xn):
            ycolors[i, j, :] = np.array(colors.to_rgba(color)) * scale[i, j]
            # xcolors[i,j,:] = xcolors[i,j,:]*min(1.5*(1-scale[i,j]),1)
            xcolors[i, j, :] = xcolors[i, j, :] * (1 - scale[i, j])

    return xcolors + ycolors


def plot_rf_prediction_and_percentiles(
    rf_prediction,
    day,
):
    """Plot SM mean prediction with 5th and 95th percentiles maps.

    Measurement locations are overlaid.
    """
    xaxis, yaxis = set_axis(rf_prediction)
    day_string = rf_prediction.input_data.days[day]

    plt.rcParams.update({"font.size": 13})
    fig, axes = plt.subplots(nrows=1, ncols=3)
    axes[0].set_aspect(1)
    axes[1].set_aspect(1)
    axes[2].set_aspect(1)
    fig.subplots_adjust(wspace=0.3)

    divider = make_axes_locatable(axes[0])
    cax = divider.append_axes("right", size="5%", pad=0.15)
    im = axes.flat[0].pcolormesh(
        xaxis,
        yaxis,
        rf_prediction.MC_mean[day],
        shading="auto",
        vmin=0.2,
        vmax=0.5,
        cmap="Spectral",
    )
    axes.flat[0].scatter(
        rf_prediction.input_data.x_crns[day_string] / 1000,
        rf_prediction.input_data.y_crns[day_string] / 1000,
        c="black",
        s=1,
    )
    axes.flat[0].set_title("Mean", fontsize=13)
    plt.colorbar(im, cax=cax)

    im = axes.flat[1].pcolormesh(
        xaxis,
        yaxis,
        rf_prediction.p5[day],
        shading="auto",
        vmin=0.2,
        vmax=0.5,
        cmap="Spectral",
    )
    axes.flat[1].set_title("5% percentile", fontsize=13)
    axes.flat[1].tick_params(axis="y", left=False, labelleft=False)
    divider = make_axes_locatable(axes[1])
    cax = divider.append_axes("right", size="5%", pad=0.15)
    plt.colorbar(im, cax=cax)

    im = axes.flat[2].pcolormesh(
        xaxis,
        yaxis,
        rf_prediction.p95[day],
        shading="auto",
        vmin=0.2,
        vmax=0.5,
        cmap="Spectral",
    )
    axes.flat[2].set_title("95% percentile", fontsize=13)
    axes.flat[2].tick_params(axis="y", left=False, labelleft=False)
    divider = make_axes_locatable(axes[2])
    cax = divider.append_axes("right", size="5%", pad=0.15)
    plt.colorbar(im, cax=cax)

    fig.suptitle(day_string)


def mc_iterations_histogram(rf_prediction, time_step, point_x, point_y):
    """Plot histogram of predicted soil moisture values after n_iters MC runs.

    point_x : point coordinate on the x axis
    point_y : point coordinate on the y axis
    """
    plt.figure()
    plt.hist(rf_prediction[:, time_step, point_x, point_y], bins=30, range=(0.15, 0.6))


# Update!
def relative_frequency_3d_section(rf_prediction, day):
    """Plot parallel 3D sections over the study area."""
    raise NotImplementedError

    if rf_prediction.apply_monte_carlo:
        nb_hist_class = 15
        sections_hist = np.empty((nb_hist_class, 4, rf_prediction.rect_geom.dim_y))
        sections_loc = np.linspace(
            10, rf_prediction.rect_geom.dim_x - 10, 4, dtype=int
        )  # 4 sections along the x-axis
        for section_id, section_coord in enumerate(sections_loc):
            for y_coord in range(rf_prediction.rect_geom.dim_y):
                sections_hist[:, section_id, y_coord], bin_edges = np.histogram(
                    rf_prediction.RF_prediction[:, day, section_coord, y_coord],
                    bins=nb_hist_class,
                    range=(0, 1),
                )

        ax = plt.figure().add_subplot(projection="3d")
        x = rf_prediction.rect_geom.grid_x[:, 0] / 1000
        y = rf_prediction.rect_geom.grid_y[0, :] / 1000
        ax.set_box_aspect(
            (np.ptp(x), np.ptp(y), 20 * np.ptp(np.linspace(0, 1.2, nb_hist_class)))
        )
        ax.set_xlim(x[0], x[-1])
        ax.set_ylim(y[0], y[-1])
        norm = mpl.colors.Normalize(
            vmin=np.min(sections_hist[np.nonzero(sections_hist)]),
            vmax=np.max(sections_hist),
        )
        mapper = mpl.cm.ScalarMappable(norm=norm, cmap="Reds")
        mapper.cmap.set_under("w", alpha=0)
        x_pos = np.linspace(x[0], x[-1], 4)
        for section_id, i in enumerate(x_pos):
            z = 0
            for class_id in range(nb_hist_class):
                zcol = np.array(
                    [
                        (mapper.to_rgba(v))
                        for v in sections_hist[class_id, section_id, :]
                    ]
                )
                ax.bar3d(i, y, z, dx=0.5, dy=0.5, dz=1.2 / nb_hist_class, color=zcol)
                z += 1 / nb_hist_class
        plt.colorbar(mapper, extend="min", shrink=0.8)
    else:
        raise RuntimeError(
            "The figure can only be plot when Monte Carlo approach is used."
        )


def comparison_station(rf_prediction):
    """Plot Rf prediction at the station location along days."""
    # TODO no hard coded path!
    station_file = (
        "c:\\users\\dega\\documents\\uq\\uq_script\\ruebeland\\sm_measurements\\"
        "Schaefertal_SoilCan_Station_daily-2021090200-2022071500.csv"
    )
    station_x = 641858.436
    station_y = 5724304.935
    station = []
    with open(station_file, "r") as st:
        lines = st.readlines()
        for row in lines[1:]:
            row_elements = row.split(",")
            station.append(float(row_elements[33]))
    nearest_node = rf_prediction.rect_geom.find_nearest_node(station_x, station_y)
    plt.figure()
    plt.plot(
        rf_prediction.RF_prediction[0, :, nearest_node[0, 0], nearest_node[0, 1]],
        label="RF predictions",
    )
    prediction_9pix_average = 0
    for x in range(-1, 2, 1):
        for y in range(-1, 2, 1):
            prediction_9pix_average += rf_prediction.RF_prediction[
                0, :, nearest_node[0, 0] + x, nearest_node[0, 1] + y
            ]
    prediction_9pix_average /= 9
    plt.plot(
        prediction_9pix_average,
        label="RF predictions - 9 pixels averaged",
    )
    plt.plot(station, label="Station")
    # plt.xticks(rf_prediction.input_data.soil_moisture_data.start_times[::30])
    plt.legend()
    output_dir = rf_prediction.working_dir
    fig_file_path = os.path.join(output_dir, "station_schafertal.tif")
    plt.savefig(fig_file_path, dpi=300)
    # matrix = np.column_stack(
    #     (rf_prediction.RF_prediction[0, :, nearest_node[0,0], nearest_node[0,1]],
    #      prediction_9pix_average,
    #      station)
    #     )
    np.savetxt(output_dir + "time_data_at_station", rf_prediction.RF_prediction[0, :, nearest_node[0,0], nearest_node[0,1]], fmt='%.4f')
